from fast_tradier.utils.YFinanceQuoteProvider import YFinanceQuoteProvider
from pathlib import Path

price_pattern_file_name = 'yahoo-price-pattern'
pattern_file = Path(Path(__file__).resolve().parent.parent, 'utils', price_pattern_file_name)
print('pattern_file_path: ', pattern_file)

yfin_quote_provider = YFinanceQuoteProvider()
ticker = ['^GSPC', 'AAPL', 'MSFT']
for t in ticker:
    price = yfin_quote_provider.get_price(t)
    print(f'ticker {t} price: {price}')


import pandas as pd
from typing import Tuple

OptionChain_Headers = ['symbol', 'strike', 'lastPrice', 'openInterest', 'ask', 'bid', 'expiration_date', 'bid_date', 'volume', 'underlying_price', 'updated_at', 'gamma', 'delta', 'vega']

class OptionUtils:
    @staticmethod
    def find_option_price(option_symbol: str, call_df: pd.DataFrame, put_df: pd.DataFrame) -> Tuple:
        '''find premium for the option_symbol, returns bid, ask and last price in tuple'''

        option_symbol = option_symbol.upper()
        iloc_idx_call = call_df.loc[call_df['symbol'] == option_symbol]
        iloc_idx_put = put_df.loc[put_df['symbol'] == option_symbol]
        target_idx = -1
        target_df = None
        result = (-1, -1, -1)

        if len(iloc_idx_call.index) > 0:
            target_idx = call_df.index.get_loc(iloc_idx_call.index[0])
            target_df = call_df
        elif len(iloc_idx_put.index) > 0:
            target_idx = put_df.index.get_loc(iloc_idx_put.index[0])
            target_df = put_df

        if target_df is not None:
            bid = target_df.iloc[target_idx]['bid']
            ask = target_df.iloc[target_idx]['ask']
            lastPrice = target_df.iloc[target_idx]['lastPrice']
            result = (bid, ask, lastPrice)
        return result


from fast_tradier.models.account.AccountOrder import AccountOrder

acc_order_json = {'id': 7650946, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'rejected', 'duration': 'day', 'price': 1.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-08-25T00:44:30.471Z', 'transaction_date': '2023-08-25T00:44:30.512Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'reason_description': 'Expiration date must be greater than the current date.', 'leg': [{'id': 7650949, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 1.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-08-25T00:44:30.471Z', 'transaction_date': '2023-08-25T00:44:30.508Z', 'class': 'option', 'option_symbol': 'SPXW230815C04510000', 'reason_description': 'Expiration date must be greater than the current date.'}, {'id': 7650952, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 1.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-08-25T00:44:30.471Z', 'transaction_date': '2023-08-25T00:44:30.511Z', 'class': 'option', 'option_symbol': 'SPXW230815C04520000', 'reason_description': 'Expiration date must be greater than the current date.'}]}
acc_order = AccountOrder(acc_order_json)
print('acc_order.price: ', acc_order.price)
print('acc_order to_json: ', acc_order.to_json())


from typing import Dict
from dataclasses import dataclass
from fast_tradier.models.ModelBase import ModelBase

@dataclass
class DataClassModelBase(ModelBase):
    def __init__(self, api_resp_dict: Dict):
        if api_resp_dict is None:
            raise Exception("api_resp_dict is None")
        self.__resp_dict = api_resp_dict
        for k, v in api_resp_dict.items():
            setattr(self, k, v)
    
    def __iter__(self):
        yield from self.__resp_dict.items()


from fast_tradier.FastTradierAsyncClient import FastTradierAsyncClient
from fast_tradier.FastTradierClient import FastTradierClient
from fast_tradier.utils.YFinanceQuoteProvider import YFinanceQuoteProvider
from fast_tradier.models.market_data.Quote import Quote
from fast_tradier.models.trading.OptionOrder import OptionLeg, OptionOrder
from fast_tradier.models.trading.EquityOrder import EquityOrder
from fast_tradier.models.trading.Sides import OptionOrderSide, EquityOrderSide
from fast_tradier.models.trading.PriceTypes import OptionPriceType, EquityPriceType
from fast_tradier.models.trading.Duration import Duration

from pathlib import Path
from datetime import datetime, timedelta
import asyncio
import json


#TODO: replace the client_id and sandbox access token with yours
sandbox_client_id = 'VA7432928'
sandbox_at = 'VuX3WA9Aq0mfORazBNM4d5bUA4zb'
yfin_real_quote_provider = YFinanceQuoteProvider()

def mock_order() -> OptionOrder:
    ticker = 'SPX'
    order_status = 'pending'
    option_symbols = ['SPXW_110223C4210', 'SPXW_110223C4220'] #TODO: replace option symbols
    sides = [OptionOrderSide.SellToOpen, OptionOrderSide.BuyToOpen]
    option_legs = []

    for i in range(len(sides)):
        opt_symbol = option_symbols[i]
        side = sides[i]
        option_legs.append(OptionLeg(underlying_symbol=ticker, option_symbol=opt_symbol, side=side, quantity=1))

    option_order = OptionOrder(ticker=ticker,
                            price=3.2,
                            price_type=OptionPriceType.Credit,
                            duration=Duration.Day,
                            option_legs=option_legs)
    return option_order

def mock_equity_order() -> EquityOrder:
    symbol = 'SPY'
    price = 379.0
    quantity = 1.0
    return EquityOrder(ticker=symbol, quantity=quantity, price=price, side=EquityOrderSide.Buy, price_type=EquityPriceType.Limit, duration=Duration.Gtc)

async def async_test():
    tasks = []
    count = 4
    tradier_client = FastTradierAsyncClient(sandbox_at, sandbox_client_id, is_prod=False, real_time_quote_provider=yfin_real_quote_provider)

    # quote1 = await tradier_client.get_quotes_async(['MSFT'])
    # print('quote1 last price: ', quote1[0].last)

    for i in range(count):
        m_order = mock_order()
        cloned_legs = m_order.clone_option_legs()
        assert cloned_legs[0].option_symbol == m_order.option_legs[0].option_symbol
        assert cloned_legs[0].side == m_order.option_legs[0].side
        tasks.append(asyncio.ensure_future(tradier_client.place_option_order_async(m_order)))

    order_ids = await asyncio.gather(*tasks)
    cancel_tasks = []
    for order_id in order_ids:
        print('order_id: ', order_id)
        cancel_tasks.append(asyncio.ensure_future(tradier_client.cancel_order_async(order_id)))

    is_canceled = await asyncio.gather(*cancel_tasks)
    for canceled in is_canceled:
        print('canceled? ', canceled)
    ### test equity order:
    equity_order = mock_equity_order()
    order_id = await tradier_client.place_equity_order_async(equity_order)
    print('equity order id: ', order_id)
    equity_order_canceled = await tradier_client.cancel_order_async(order_id)
    print('equity order canceld? ', equity_order_canceled)
    ### get option chain for spx
    ticker = 'spx'
    expiration = '2023-08-31' #TODO: replace the expiration date
    opt_chain_result = await tradier_client.get_option_chain_async(symbol=ticker, expiration=expiration)
    print('result of option chain: ', opt_chain_result)
    positions = await tradier_client.get_positions_async()
    print('positions: ', positions)
    exps = await tradier_client.get_option_expirations_async(symbol=ticker)
    print(f'ticker: {ticker} has exps: {exps}')

    print('------' * 10)
    balances = await tradier_client.get_account_balance_async()
    print('balances: ', balances.total_cash)
    account_orders = await tradier_client.get_account_orders_async()
    print('account orders count: ', len(account_orders))
    orders_json = []
    for acc in account_orders:
        print('acc: ', acc.to_json())
        orders_json.append(acc.to_json())
    json_file = Path(Path(__file__).resolve().parent, "orders_json.json")
    with open(json_file, 'w') as fp:
        fp.write(json.dumps(orders_json))

    if hasattr(balances, 'dt') and balances.pdt is not None:
        print('balances.pdt.to_json(): ', balances.pdt.to_json())
    elif hasattr(balances, 'margin') and balances.margin is not None:
        print('balances.margin.to_json: ', balances.margin.to_json())
    elif hasattr(balances, 'cash') and balances.cash is not None:
        print('balances.cash.to_json: ', balances.cash.to_json())

async def real_test_async():
    prod_at = 'OLtA2pZ3A9FFY7ZIsKcfcKRhJYAd'
    client_id = '6YA19668'
    tradier_client = FastTradierAsyncClient(prod_at, client_id, is_prod=True, real_time_quote_provider=yfin_real_quote_provider)
    expiration = '2023-10-20'
    ticker = 'spx'
    result = await tradier_client.get_option_chain_async(symbol=ticker, expiration=expiration)
    call_df, put_df = result['call_chain'], result['put_chain']
    call_df.to_csv('call_df.csv')
    put_df.to_csv('put_df.csv')
    print(call_df.head())
    print('------')
    print(put_df.head())

def sync_test():
    count = 4
    tradier_client = FastTradierClient(sandbox_at, sandbox_client_id, is_prod=False, real_time_quote_provider=yfin_real_quote_provider)
    # quote1 = tradier_client.get_quotes(['MSFT'])
    # print('quote1 last price: ', quote1[0].last)
    
    # m_order = mock_order()
    # cloned_legs = m_order.clone_option_legs()
    # assert cloned_legs[0].option_symbol == m_order.option_legs[0].option_symbol
    # assert cloned_legs[0].side == m_order.option_legs[0].side

    # order_id = tradier_client.place_option_order(m_order)
    # print('option order id: ', order_id)
    # order_status = tradier_client.get_order_status(order_id=order_id)
    # print(f'order_status: {order_status} for order_id: {order_id}')
    # canceled_order = tradier_client.cancel_order(order_id)
    # print('canceled order? ', canceled_order)
    
    # print('-------' * 10)
    # ### test equity order:
    # equity_order = mock_equity_order()
    # order_id = tradier_client.place_equity_order(equity_order)
    # print('equity order id: ', order_id)
    # equity_order_canceled = tradier_client.cancel_order(order_id)
    # print('equity order canceld? ', equity_order_canceled)

    # ### get option chain for spx
    # ticker = 'spx'
    # expiration = '2023-08-31' #TODO: replace the expiration date
    # opt_chain_result = tradier_client.get_option_chain(symbol=ticker, expiration=expiration)
    # print('result of option chain: ', opt_chain_result)
    # exps = tradier_client.get_option_expirations(symbol=ticker)
    # print(f'ticker: {ticker} has exps: {exps}')
    # positions = tradier_client.get_positions()
    # print('positions: ', positions)

    print('------' * 10)
    balances = tradier_client.get_account_balance()
    print('balances: ', balances.total_cash)
    if hasattr(balances, 'dt') and balances.pdt is not None:
        print('balances.pdt.to_json(): ', balances.pdt.to_json())
    elif hasattr(balances, 'margin') and balances.margin is not None:
        print('balances.margin.to_json: ', balances.margin.to_json())
    elif hasattr(balances, 'cash') and balances.cash is not None:
        print('balances.cash.to_json: ', balances.cash.to_json())

def get_history():
    ticker = 'SPY'
    start_date = datetime.now() + timedelta(days=-30)
    end_date = datetime.now() + timedelta(days=-2)
    # print(f'start_date: {start_date.date()}')
    # tradier_client = FastTradierClient(sandbox_at, sandbox_client_id, is_prod=False, real_time_quote_provider=yfin_real_quote_provider)
    # result_df = tradier_client.get_history(symbol=ticker, start_date=start_date.date(), end_date=end_date.date())
    # print('result_df: ', result_df)

async def get_history_async():
    ticker = 'SPY'
    start_date = datetime.now() + timedelta(days=-30)
    end_date = datetime.now() + timedelta(days=-2)
    tradier_client = FastTradierAsyncClient(sandbox_at, sandbox_client_id, is_prod=False, real_time_quote_provider=yfin_real_quote_provider)
    result_df = await tradier_client.get_history_async(symbol=ticker, start_date=start_date, end_date=end_date)
    print('result_df: ', result_df)

def get_order():
    order_id = '9389363'
    tradier_client = FastTradierClient(sandbox_at, sandbox_client_id, is_prod=False, real_time_quote_provider=yfin_real_quote_provider)
    cur_orders = tradier_client.get_account_orders()
    for cur_o in cur_orders:
        print(f'cur_o: {cur_o.to_json()}')

'''
cur_o: {'id': 9389363, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'filled', 'duration': 'day', 'price': 1.8, 'avg_fill_price': -2.25, 'exec_quantity': 2.0, 'last_fill_price': 0.0, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T15:44:06.399Z', 'transaction_date': '2023-12-07T15:44:06.399Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'leg': [{'id': 9389364, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 1.8, 'avg_fill_price': 3.5, 'exec_quantity': 1.0, 'last_fill_price': 3.5, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T15:44:06.399Z', 'transaction_date': '2023-12-07T15:44:06.399Z', 'class': 'option', 'option_symbol': 'SPXW231207C04590000'}, {'id': 9389365, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 1.8, 'avg_fill_price': 1.25, 'exec_quantity': 1.0, 'last_fill_price': 1.25, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T15:44:06.399Z', 'transaction_date': '2023-12-07T15:44:06.399Z', 'class': 'option', 'option_symbol': 'SPXW231207C04600000'}]}
cur_o: {'id': 9389401, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'filled', 'duration': 'day', 'price': 1.8, 'avg_fill_price': -2.45, 'exec_quantity': 2.0, 'last_fill_price': 0.0, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T15:46:04.473Z', 'transaction_date': '2023-12-07T15:46:04.473Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'leg': [{'id': 9389402, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 1.8, 'avg_fill_price': 3.9, 'exec_quantity': 1.0, 'last_fill_price': 3.9, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T15:46:04.473Z', 'transaction_date': '2023-12-07T15:46:04.473Z', 'class': 'option', 'option_symbol': 'SPXW231207C04590000'}, {'id': 9389403, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 1.8, 'avg_fill_price': 1.45, 'exec_quantity': 1.0, 'last_fill_price': 1.45, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T15:46:04.473Z', 'transaction_date': '2023-12-07T15:46:04.473Z', 'class': 'option', 'option_symbol': 'SPXW231207C04600000'}]}
cur_o: {'id': 9390234, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'canceled', 'duration': 'day', 'price': 0.8, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T16:20:04.707Z', 'transaction_date': '2023-12-07T16:24:04.813Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'leg': [{'id': 9390235, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'canceled', 'duration': 'day', 'price': 0.8, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T16:20:04.707Z', 'transaction_date': '2023-12-07T16:24:04.813Z', 'class': 'option', 'option_symbol': 'SPXW231207P04545000'}, {'id': 9390236, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'canceled', 'duration': 'day', 'price': 0.8, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T16:20:04.707Z', 'transaction_date': '2023-12-07T16:24:04.813Z', 'class': 'option', 'option_symbol': 'SPXW231207P04535000'}]}
cur_o: {'id': 9395790, 'type': 'debit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'filled', 'duration': 'day', 'price': 0.45, 'avg_fill_price': 0.45, 'exec_quantity': 2.0, 'last_fill_price': 0.0, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:34:05.939Z', 'transaction_date': '2023-12-07T20:37:13.533Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'leg': [{'id': 9395791, 'type': 'debit', 'symbol': 'SPX', 'side': 'buy_to_close', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 0.45, 'avg_fill_price': 0.5, 'exec_quantity': 1.0, 'last_fill_price': 0.5, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:34:05.939Z', 'transaction_date': '2023-12-07T20:37:13.533Z', 'class': 'option', 'option_symbol': 'SPXW231207C04590000'}, {'id': 9395792, 'type': 'debit', 'symbol': 'SPX', 'side': 'sell_to_close', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 0.45, 'avg_fill_price': 0.05, 'exec_quantity': 1.0, 'last_fill_price': 0.05, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:34:05.939Z', 'transaction_date': '2023-12-07T20:37:13.533Z', 'class': 'option', 'option_symbol': 'SPXW231207C04600000'}]}
cur_o: {'id': 9395793, 'type': 'debit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'filled', 'duration': 'day', 'price': 0.45, 'avg_fill_price': 0.45, 'exec_quantity': 2.0, 'last_fill_price': 0.0, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:34:07.018Z', 'transaction_date': '2023-12-07T20:37:03.361Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'leg': [{'id': 9395794, 'type': 'debit', 'symbol': 'SPX', 'side': 'buy_to_close', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 0.45, 'avg_fill_price': 0.5, 'exec_quantity': 1.0, 'last_fill_price': 0.5, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:34:07.018Z', 'transaction_date': '2023-12-07T20:37:03.361Z', 'class': 'option', 'option_symbol': 'SPXW231207C04590000'}, {'id': 9395795, 'type': 'debit', 'symbol': 'SPX', 'side': 'sell_to_close', 'quantity': 1.0, 'status': 'filled', 'duration': 'day', 'price': 0.45, 'avg_fill_price': 0.05, 'exec_quantity': 1.0, 'last_fill_price': 0.05, 'last_fill_quantity': 1.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:34:07.018Z', 'transaction_date': '2023-12-07T20:37:03.361Z', 'class': 'option', 'option_symbol': 'SPXW231207C04600000'}]}
cur_o: {'id': 9396783, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.476Z', 'transaction_date': '2023-12-07T20:52:45.550Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000', 'leg': [{'id': 9396787, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.476Z', 'transaction_date': '2023-12-07T20:52:45.546Z', 'class': 'option', 'option_symbol': 'SPXW231102C04210000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}, {'id': 9396792, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.476Z', 'transaction_date': '2023-12-07T20:52:45.548Z', 'class': 'option', 'option_symbol': 'SPXW231102C04220000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}]}
cur_o: {'id': 9396784, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.465Z', 'transaction_date': '2023-12-07T20:52:45.550Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000', 'leg': [{'id': 9396788, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.465Z', 'transaction_date': '2023-12-07T20:52:45.546Z', 'class': 'option', 'option_symbol': 'SPXW231102C04210000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}, {'id': 9396791, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.465Z', 'transaction_date': '2023-12-07T20:52:45.548Z', 'class': 'option', 'option_symbol': 'SPXW231102C04220000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}]}
cur_o: {'id': 9396785, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.477Z', 'transaction_date': '2023-12-07T20:52:45.550Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000', 'leg': [{'id': 9396789, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.477Z', 'transaction_date': '2023-12-07T20:52:45.546Z', 'class': 'option', 'option_symbol': 'SPXW231102C04210000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}, {'id': 9396793, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.477Z', 'transaction_date': '2023-12-07T20:52:45.549Z', 'class': 'option', 'option_symbol': 'SPXW231102C04220000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}]}
cur_o: {'id': 9396782, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy', 'quantity': 2.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.465Z', 'transaction_date': '2023-12-07T20:52:45.551Z', 'class': 'multileg', 'num_legs': 2, 'strategy': 'spread', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000', 'leg': [{'id': 9396786, 'type': 'credit', 'symbol': 'SPX', 'side': 'sell_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.465Z', 'transaction_date': '2023-12-07T20:52:45.547Z', 'class': 'option', 'option_symbol': 'SPXW231102C04210000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}, {'id': 9396790, 'type': 'credit', 'symbol': 'SPX', 'side': 'buy_to_open', 'quantity': 1.0, 'status': 'rejected', 'duration': 'day', 'price': 3.2, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:45.465Z', 'transaction_date': '2023-12-07T20:52:45.549Z', 'class': 'option', 'option_symbol': 'SPXW231102C04220000', 'reason_description': 'There is no price. Security symbol: SPXW231102C04210000'}]}
cur_o: {'id': 9396797, 'type': 'limit', 'symbol': 'SPY', 'side': 'buy', 'quantity': 1.0, 'status': 'canceled', 'duration': 'gtc', 'price': 379.0, 'avg_fill_price': 0.0, 'exec_quantity': 0.0, 'last_fill_price': 0.0, 'last_fill_quantity': 0.0, 'remaining_quantity': 0.0, 'create_date': '2023-12-07T20:52:46.371Z', 'transaction_date': '2023-12-07T20:52:46.788Z', 'class': 'equity'}

'''

# asyncio.run(async_test())
# asyncio.run(real_test_async())
print('-------finished async tests--------')
# sync_test()
print('-------finished sync tests-------')
# get_order()
# get_history()
asyncio.run(get_history_async())


from datetime import datetime, date, timedelta
from typing import Optional
import pandas as pd
import pytz

US_Eastern_TZ = pytz.timezone('US/Eastern')
YMD_Format = "%Y-%m-%d"
YYYYMDHHMM_Format = "YYYY-M-D HH:mm"
YMDHMS_Format = "%Y-%m-%d %H:%M:%S"

class TimeUtils:

    @staticmethod
    def convert_unix_ts(ts: int, tz: pytz.timezone = US_Eastern_TZ) -> Optional[datetime]:
        '''convert unix timestamp to datetime with the given timezone, US East by default'''
        if ts is None:
            return None

        if isinstance(ts, pd.Timestamp):
            ts = ts.to_pydatetime().timestamp()

        while ts > 1e10:
            ts = ts/1000
        return datetime.fromtimestamp(ts, tz)

    @staticmethod
    def us_east_now() -> datetime:
        return datetime.now().astimezone(US_Eastern_TZ)

    @staticmethod
    def parse_day_str(date_str: str, timezone: pytz.timezone = US_Eastern_TZ) -> date:
        """Parse a date string to date object in the intended timezone (US Eastern default).
        example input: '2023-12-10'
        """
        day_dt:datetime = datetime.strptime(date_str, YMD_Format).astimezone(timezone)
        return day_dt.date()
    
    @staticmethod
    def today_date(timezone: pytz.timezone = US_Eastern_TZ) -> date:
        today = datetime.now().astimezone(timezone)
        return today.date()
    
    @staticmethod
    def past_date(days_ago: int, timezone: pytz.timezone = US_Eastern_TZ) -> date:
        """Get a past date x days before today's date."""
        today_dt = TimeUtils.today_date(timezone)
        days_ago = days_ago if days_ago > 0 else days_ago
        past_dt = today_dt + timedelta(days=-days_ago)
        return past_dt

    @staticmethod
    def today_str(timezone: pytz.timezone = US_Eastern_TZ) -> str:
        today_dt = TimeUtils.today_date(timezone)
        return today_dt.strftime(YMD_Format)
        


from __future__ import annotations
from interface import implements
from typing import Dict, Optional

from abc import abstractmethod
import json
import jsonpickle

from fast_tradier.interfaces.IModel import IModel

class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        return obj.serialize()

class ModelBase(implements(IModel)):

    def __init__(self):
        raise NotImplementedError("ModelBase cannot be initiated")

    @abstractmethod
    def __iter__(self):
        '''method rquired for to_json() method. Optional if to_json() is not needed'''
        raise NotImplemented("__iter__() not implemented in ModelBase")

    def __str__(self):
        return json.dumps(dict(self), ensure_ascii=False)
    
    def __repr__(self) -> str:
        return self.__str__()

    def to_json(self) -> Dict:
        return json.loads(self.__str__())

    def serialize(self) -> Dict:
        '''serialize class object to json, which is parsable back to class object'''
        return jsonpickle.encode(self)

    @classmethod
    def deserialize_from_json(cls, json_obj) -> Optional[ModelBase]:
        '''Parse json object to class object. Could throw exception if not parsable'''
        return jsonpickle.decode(json_obj)


import os

from interface import implements
from typing import Optional
from autoscraper import AutoScraper
from pathlib import Path

from fast_tradier.interfaces.IRealTimeQuoteProvider import IRealTimeQuoteProvider

class YFinanceQuoteProvider(implements(IRealTimeQuoteProvider)):
    '''Fetch real time stock quote from Yahoo Finance using web scraper'''    
    def __init__(self, price_pattern_file_path: Optional[str] = None) -> None:
        if price_pattern_file_path is not None and not os.path.exists(price_pattern_file_path):
            raise FileNotFoundError(f'price pattern file {price_pattern_file_path} does not exist')
        default_price_pattern_file = Path(Path(__file__).resolve().parent, 'yahoo-price-pattern')
        self.__pattern_file_path = default_price_pattern_file if price_pattern_file_path is None else price_pattern_file_path
        self.__scraper = AutoScraper()
        self.__scraper.load(self.__pattern_file_path)
        self.__yfinance_base_url = 'https://finance.yahoo.com/quote/{symbol}'

    def get_price(self, symbol: str) -> Optional[float]:
        try:
            url = self.__yfinance_base_url.format(symbol = YFinanceQuoteProvider.convert_to_yf_ticker(symbol))
            result = self.__scraper.get_result_similar(url)
            quote_price = None
            if result is not None and len(result) > 0:
                quote_price = float(result[0])
            return quote_price
        except Exception as ex:
            return None

    @classmethod
    def convert_to_yf_ticker(cls, ticker: str):
        '''convert ticker/symbol to yahoo finance ticker'''
        ticker_yahoo_map = {
            'SPX' : '^GSPC',
            'SPXW' : '^GSPC',
            '$SPX.X' : '^GSPC',
            '$VIX.X' : '^VIX',
            'VIX' : '^VIX',
            'VVIX' : '^VVIX'}

        cur_ticker = ticker.upper()
        if cur_ticker in ticker_yahoo_map:
            return ticker_yahoo_map[cur_ticker]

        return ticker


from fast_tradier.models.ModelBase import ModelBase
from typing import List, Dict, Optional
from fast_tradier.models.trading.Sides import EquityOrderSide
from fast_tradier.models.trading.PriceTypes import EquityPriceType
from fast_tradier.models.trading.Duration import Duration
from fast_tradier.models.trading.OrderBase import OrderBase

class EquityOrder(OrderBase):
    def __init__(self, ticker: str, quantity: float, price: float, side: EquityOrderSide, price_type: EquityPriceType, duration: Duration, stop: Optional[float] = None) -> None:
        self.__ticker = ticker
        self.__quantity = quantity
        self.__price = price
        self.__side = side.value
        self.__price_type = price_type.value
        self.__duration = duration.value
        self.__stop = stop
        self.__status = None
        self.__order_class = 'equity'
        self.__id = None

    @property
    def ticker(self) -> str:
        return self.__ticker
    
    @property
    def quantity(self) -> float:
        return self.__quantity
    
    @quantity.setter
    def quantity(self, new_value: float):
        self.__quantity = new_value

    @property
    def price(self) -> float:
        return self.__price

    @price.setter
    def price(self, new_value : float):
        self.__price = new_value

    @property
    def side(self) -> str:
        return self.__side

    @property
    def price_type(self) -> str:
        return self.__price_type

    @property
    def duration(self) -> str:
        return self.__duration

    @property
    def status(self) -> str:
        return self.__status

    @status.setter
    def status(self, new_value: str):
        self.__status = new_value

    @property
    def id(self) -> int:
        return self.__id

    @id.setter
    def id(self, new_value: int):
        self.__id = new_value

    @property
    def order_class(self) -> str:
        return self.__order_class

    @property
    def stop(self) -> Optional[float]:
        return self.__stop

    def __iter__(self):
        result = {
            "id": self.id,
            "status": self.status,
            "side": self.side,
            "symbol": self.ticker,
            "quantity": self.quantity,
            "duration": self.duration,
            "price": self.price,
            "type": self.price_type,
            "class": self.order_class,
            "stop": self.stop
        }

        yield from result.items()


from typing import Dict, List

from fast_tradier.models.trading.TOSTradierConverter import TOSTradierConverter
from fast_tradier.models.ModelBase import ModelBase
from fast_tradier.models.trading.Sides import OptionOrderSide
from fast_tradier.models.trading.PriceTypes import OptionPriceType
from fast_tradier.models.trading.Duration import Duration
from fast_tradier.models.trading.OrderBase import OrderBase

class OptionLeg(OrderBase):
    def __init__(self, underlying_symbol: str, option_symbol: str, side: OptionOrderSide, quantity: int, convert_tos_symbol: bool = True) -> None:
        self.__underlying_symbol = underlying_symbol
        self.__tos_option_symbol = TOSTradierConverter.tradier_to_tos(option_symbol) if convert_tos_symbol else option_symbol
        self.__option_symbol = TOSTradierConverter.tos_to_tradier(option_symbol) if convert_tos_symbol else option_symbol
        self.__side = side.value
        self.__quantity = quantity
        self.__convert_tos_symbol = convert_tos_symbol

    @property
    def convert_tos_symbol(self) -> bool:
        return self.__convert_tos_symbol

    @property
    def underlying_symbol(self) -> str:
        return self.__underlying_symbol

    @property
    def option_symbol(self) -> str:
        return self.__option_symbol

    @property
    def tos_option_symbol(self) -> str:
        return self.__tos_option_symbol

    @property
    def side(self) -> str:
        return self.__side

    @property
    def quantity(self) -> int:
        return self.__quantity

    # reverse open to close or vice versa
    def reverse_side(self) -> None:
        '''reverse option side: from sell_to_open to buy_to_close or buy_to_open to sell_to_close'''
        if self.__side == 'buy_to_open':
            self.__side = 'sell_to_close'
        elif self.__side == 'sell_to_open':
            self.__side = 'buy_to_close'
    
class OptionOrder(ModelBase):
    def __init__(self, ticker: str, price: float, price_type: OptionPriceType, duration: Duration, option_legs: List[OptionLeg]):
        if option_legs is None or len(option_legs) == 0:
            raise Exception('option_legs must contain at least 1 leg')

        self.__ticker = ticker
        self.__price = price
        self.__price_type = price_type.value
        self.__duration = duration.value
        self.__option_legs = option_legs
        self.__status = 'pending'
        self.__order_class = 'multileg' if len(option_legs) > 1 else 'option'
        self.__id = None

    @property
    def ticker(self) -> str:
        return self.__ticker

    @ticker.setter
    def ticker(self, new_value: str):
        self.__ticker = new_value

    @property
    def price(self) -> float:
        return self.__price

    @property
    def price_type(self) -> str:
        return self.__price_type

    @property
    def duration(self) -> str:
        return self.__duration
    
    @property
    def option_legs(self) -> List[OptionLeg]:
        return self.__option_legs

    @price.setter
    def price(self, new_value : float):
        self.__price = new_value

    @property
    def status(self) -> str:
        return self.__status

    @status.setter
    def status(self, new_value: str):
        self.__status = new_value

    @property
    def id(self) -> int:
        return self.__id

    @id.setter
    def id(self, new_value: int):
        self.__id = new_value

    @property
    def order_class(self) -> str:
        return self.__order_class
    
    def clone_option_legs(self, reverse_side: bool = False) -> List[OptionLeg]:
        '''deep clone option_legs'''
        cloned_legs = []
        for opt_leg in self.option_legs:
            leg = OptionLeg(
                underlying_symbol=opt_leg.option_symbol,
                option_symbol=opt_leg.option_symbol,
                side=OptionOrderSide(opt_leg.side),
                quantity=opt_leg.quantity,
                convert_tos_symbol=opt_leg.convert_tos_symbol)

            if reverse_side:
                leg.reverse_side()

            cloned_legs.append(leg)

        return cloned_legs
    
    def __iter__(self):
        '''for preparing to_json(), which will be sent to Tradier API to place option order'''
        result = {
            "id": self.id,
            "status": self.status,
            "symbol": self.ticker,
            "duration": self.duration,
            "price": self.price,
            "type": self.price_type,
            "class": self.order_class,
        }
        if len(self.option_legs) == 1:
            result['option_symbol'] = self.__option_legs[0].option_symbol
            result['tos_option_symbol'] = self.__option_legs[0].tos_option_symbol
            result['side'] = self.__option_legs[0].side
            result['quantity'] = self.__option_legs[0].quantity
        elif len(self.option_legs) > 1:
            for i in range(len(self.__option_legs)):
                opt_item = self.__option_legs[i]
                symbol_key = f'option_symbol[{i}]'
                result[symbol_key] = opt_item.option_symbol
                tos_symbol_key = f'tos_option_symbol[{i}]'
                result[tos_symbol_key] = opt_item.tos_option_symbol
                side_key = f'side[{i}]'
                result[side_key] = f'{opt_item.side}'
                quant_key = f'quantity[{i}]'
                result[quant_key] = f'{opt_item.quantity}'

        yield from result.items()


# to run pytest:
# python -m pytest tests/test_client.py
import httpx
import pytest
import json
import arrow
import pytz
import random
from datetime import datetime, timedelta
from typing import Dict
from pathlib import Path
from pytest_httpx import HTTPXMock
from fast_tradier.utils.TimeUtils import TimeUtils
from fast_tradier.models.account import AccountBalance
from fast_tradier.FastTradierClient import FastTradierClient
from fast_tradier.FastTradierAsyncClient import FastTradierAsyncClient
from fast_tradier.utils.OptionUtils import OptionUtils

from fast_tradier.models.trading.OptionOrder import OptionLeg, OptionOrder
from fast_tradier.models.trading.EquityOrder import EquityOrder
from fast_tradier.models.trading.Sides import OptionOrderSide, EquityOrderSide
from fast_tradier.models.trading.PriceTypes import OptionPriceType, EquityPriceType
from fast_tradier.models.trading.Duration import Duration

eastern = pytz.timezone('US/Eastern')
mock_response_file_name = 'tradier_api_response_mock.json'
account_id = 'abc'
account_at = '123xyz'

def mock_order() -> OptionOrder:
    ticker = 'SPX'
    order_status = 'pending'
    option_symbols = ['SPXW_080823C4510', 'SPXW_080823C4520'] #TODO: replace option symbols
    sides = [OptionOrderSide.SellToOpen, OptionOrderSide.BuyToOpen]
    option_legs = []

    for i in range(len(sides)):
        opt_symbol = option_symbols[i]
        side = sides[i]
        option_legs.append(OptionLeg(underlying_symbol=ticker, option_symbol=opt_symbol, side=side, quantity=1))

    option_order = OptionOrder(ticker=ticker,
                            price=1.2,
                            price_type=OptionPriceType.Credit,
                            duration=Duration.Day,
                            option_legs=option_legs)
    return option_order

def get_mock_response() -> Dict:
    mock_file_path = Path(Path(__file__).resolve().parent, mock_response_file_name)
    with open(mock_file_path, 'r') as json_file:
        return json.load(json_file)

def test_client_init():
    client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False)
    assert client is not None

@pytest.mark.asyncio
async def test_option_expirations_call(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    httpx_mock.add_response(json=mock_resp['option_expirations_resp'])

    symbol = 'SPX'
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        results = tradier_client.get_option_expirations(symbol=symbol)
        assert len(results) > 0
        assert results[0] == "2023-08-08"
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        results = await tradier_client.get_option_expirations_async(symbol=symbol)
        assert len(results) > 0
        assert results[0] == "2023-08-08"

@pytest.mark.asyncio
async def test_is_market_open_today(httpx_mock: HTTPXMock):
    today = datetime.now().astimezone(eastern)
    today_str = today.strftime("%Y-%m-%d")
    now_t = arrow.get(f'{today_str} 12:00', 'YYYY-M-D HH:mm', tzinfo=eastern)
    now = now_t.datetime
    tomorrow = now + timedelta(days=1)
    year = now.year
    month = now.month

    mock_json_resp = {
        "calendar": {
            "month": month,
            "year": year,
            "days": {
                "day": []
            }
        }
    }

    days = [now, tomorrow]
    for t in days:
        mock_json_resp["calendar"]["days"]["day"].append({
            "date": f'{t.year}-{t.month}-{t.day}',
            "status": "open",
            "open": {
                "start": "09:30",
                "end": "16:00"
            }
        })
    httpx_mock.add_response(json=mock_json_resp)
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        is_open, day_arr, today_open_window = tradier_client.is_market_open_today()
        assert len(day_arr) == len(days)
        assert day_arr[0] == f'{days[0].year}-{days[0].month}-{days[0].day}'
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        is_open, day_arr, today_open_window = await tradier_client.is_market_open_today_async()
        assert len(day_arr) == len(days)
        assert day_arr[0] == f'{days[0].year}-{days[0].month}-{days[0].day}'

@pytest.mark.asyncio
async def test_get_quotes(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    httpx_mock.add_response(json=mock_resp['quotes_resp'])
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        tickers = ['AAPL']
        quotes_list = tradier_client.get_quotes(symbols=tickers)
        assert len(quotes_list) > 0
        assert quotes_list[0].symbol == 'AAPL'
        assert quotes_list[0].type == 'stock'
        assert quotes_list[0].ask == 208.21
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        tickers = ['AAPL']
        quotes_list = await tradier_client.get_quotes_async(symbols=tickers)
        assert len(quotes_list) > 0
        assert quotes_list[0].symbol == 'AAPL'
        assert quotes_list[0].type == 'stock'
        assert quotes_list[0].ask == 208.21

@pytest.mark.asyncio
async def test_get_order_status(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    httpx_mock.add_response(json=mock_resp['orders_resp'])
    target_order_id = 228175
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        order_status = tradier_client.get_order_status(order_id=target_order_id)
        assert order_status == 'expired'
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        order_status = await tradier_client.get_order_status_async(order_id=target_order_id)
        assert order_status == 'expired'
@pytest.mark.asyncio
async def test_get_option_chain(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    httpx_mock.add_response(json=mock_resp["option_chain_resp"])
    expiration = '2023-08-08'
    symbol = 'VIX'
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        result = tradier_client.get_option_chain(symbol=symbol, expiration=expiration)
        assert result["expiration"] == expiration
        assert result["ticker"] == symbol
        assert len(result["call_chain"]) > 0
        assert len(result["put_chain"]) > 0
        call_df, put_df = result["call_chain"], result["put_chain"]
        bid, ask, lastPrice = OptionUtils.find_option_price(option_symbol='VXX190517C00016000', call_df=call_df, put_df=put_df)
        assert bid == 10.85
        assert ask == 11.0
        assert lastPrice == 0
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        result = await tradier_client.get_option_chain_async(symbol=symbol, expiration=expiration)
        assert result["expiration"] == expiration
        assert result["ticker"] == symbol
        assert len(result["call_chain"]) > 0
        assert len(result["put_chain"]) > 0
        call_df, put_df = result["call_chain"], result["put_chain"]
        bid, ask, lastPrice = OptionUtils.find_option_price(option_symbol='VXX190517C00016000', call_df=call_df, put_df=put_df)
        assert bid == 10.85
        assert ask == 11.0
        assert lastPrice == 0

@pytest.mark.asyncio
async def test_get_account_balance(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    httpx_mock.add_response(json=mock_resp["account_balances_resp"])
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        account_balance = tradier_client.get_account_balance()
        assert account_balance is not None
        assert account_balance.account_type == 'margin'
        assert account_balance.margin is not None
        assert account_balance.margin.option_buying_power == 6363.86
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        account_balance = await tradier_client.get_account_balance_async()
        assert account_balance is not None
        assert account_balance.account_type == 'margin'
        assert account_balance.margin is not None
        assert account_balance.margin.option_buying_power == 6363.86

@pytest.mark.asyncio
async def test_get_account_positions(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    httpx_mock.add_response(json=mock_resp["positions_resp"])
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        positions = tradier_client.get_positions()
        assert len(positions) > 0
        assert positions[0].symbol == 'AAPL'
        assert positions[0].cost_basis == 207.01
        assert positions[1].symbol == 'AMZN'
        assert positions[3].symbol == 'FB'
        assert positions[3].cost_basis == 173.04
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        positions = await tradier_client.get_positions_async()
        assert len(positions) > 0
        assert positions[0].symbol == 'AAPL'
        assert positions[0].cost_basis == 207.01
        assert positions[1].symbol == 'AMZN'
        assert positions[3].symbol == 'FB'
        assert positions[3].cost_basis == 173.04

@pytest.mark.asyncio
async def test_get_account_orders(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    httpx_mock.add_response(json=mock_resp["orders_resp"])
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        orders = tradier_client.get_account_orders()
        assert len(orders) == 3
        assert orders[0].symbol == 'AAPL'
        assert orders[0].id == 228175
        assert orders[1].symbol == 'SPY'
        assert len(orders[2].leg) > 0
        assert orders[2].leg[0].id == 229064
        assert orders[2].leg[1].option_symbol == 'SPY180720C00274000'
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        orders = await tradier_client.get_account_orders_async()
        assert len(orders) == 3
        assert orders[0].symbol == 'AAPL'
        assert orders[0].id == 228175
        assert orders[1].symbol == 'SPY'
        assert len(orders[2].leg) > 0
        assert orders[2].leg[0].id == 229064
        assert orders[2].leg[1].option_symbol == 'SPY180720C00274000'

@pytest.mark.asyncio
async def test_place_order(httpx_mock: HTTPXMock):
    mock_order_id = random.randint(100, 200)
    httpx_mock.add_response(json={
        "order": {
            "status": "ok",
            "id": mock_order_id,
        }
    })

    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token=account_at, account_id=account_id, is_prod=False, http_client=client)
        order_id = tradier_client.place_option_order(mock_order())
        assert order_id == mock_order_id
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        order_id = await tradier_client.place_option_order_async(mock_order())
        assert order_id == mock_order_id

@pytest.mark.asyncio
async def test_cancel_order(httpx_mock: HTTPXMock):
    mock_order_id = random.randint(100, 200)
    httpx_mock.add_response(json={
        "order": {
            "status": "ok",
            "id": mock_order_id,
        }
    })

    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token=account_at, account_id=account_id, is_prod=False, http_client=client)
        is_canceled = tradier_client.cancel_order(mock_order_id)
        assert is_canceled

    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        is_canceled = await tradier_client.cancel_order_async(mock_order_id)
        assert is_canceled

@pytest.mark.asyncio
async def test_modify_option_order(httpx_mock: HTTPXMock):
    mock_order_id = random.randint(100, 200)
    httpx_mock.add_response(json={
        "order": {
            "status": "ok",
            "id": mock_order_id,
        }
    })
    
    order = mock_order()
    order.id = mock_order_id

    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token=account_at, account_id=account_id, is_prod=False, http_client=client)
        is_modified = tradier_client.modify_option_order(order)
        assert is_modified
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        is_modified = await tradier_client.modify_option_order_async(order)
        assert is_modified

@pytest.mark.asyncio
async def test_get_history_quotes(httpx_mock: HTTPXMock):
    mock_resp = get_mock_response()
    symbol = 'AAPL'
    start_date = TimeUtils.past_date(days_ago=100)
    end_date = TimeUtils.today_date()

    httpx_mock.add_response(json=mock_resp["history_resp"])
    with httpx.Client() as client:
        tradier_client = FastTradierClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        history_df = tradier_client.get_history(symbol=symbol, start_date=start_date, end_date=end_date)
        assert history_df is not None
        assert len(list(history_df.columns)) > 0
        assert history_df.shape[0] == 3
        assert history_df.shape[1] == 5
        assert history_df.iloc[-1]['Close'] == 148.26
    
    # async test:
    async with httpx.AsyncClient() as client:
        tradier_client = FastTradierAsyncClient(access_token = account_at, account_id=account_id, is_prod=False, http_client=client)
        history_df2 = await tradier_client.get_history_async(symbol=symbol, start_date=start_date, end_date=end_date)
        assert history_df2 is not None
        assert len(list(history_df2.columns)) > 0
        assert history_df2.shape[0] == 3
        assert history_df2.shape[1] == 5
        assert history_df2.iloc[-1]['Close'] == 148.26

def test_timeutils():
    today_str = TimeUtils.today_str()
    assert today_str is not None
    parsed_today_dt = TimeUtils.parse_day_str(today_str)
    assert parsed_today_dt == TimeUtils.today_date()
    yesterday_dt = TimeUtils.past_date(days_ago=1)
    assert yesterday_dt is not None
    assert yesterday_dt == TimeUtils.today_date() + timedelta(days=-1)


import json
from dataclasses import dataclass
from typing import Dict, Optional, List

from fast_tradier.models.DataClassModelBase import DataClassModelBase

@dataclass
class Leg(DataClassModelBase):
    id: int
    type: str
    symbol: str
    side: str
    quantity: float
    status: str
    duration: str
    price: float
    avg_fill_price: float
    exec_quantity: float
    last_fill_price: float
    last_fill_quantity: float
    remaining_quantity: float
    create_date: str
    transaction_date: str
    # class: str
    class_type: str
    option_symbol: Optional[str]
    def __init__(self, api_resp_dict: Dict):
        super().__init__(api_resp_dict)
        if 'class' in api_resp_dict:
            self.class_type = api_resp_dict['class']

@dataclass
class AccountOrder(DataClassModelBase):
    id: int
    type: str
    symbol: str
    side: str
    quantity: float
    status: str
    duration: str
    price: float
    avg_fill_price: float
    exec_quantity: float
    last_fill_price: float
    last_fill_quantity: float
    remaining_quantity: float
    create_date: str
    transaction_date: str
    class_type: str
    num_legs: Optional[int]
    strategy: Optional[str]
    leg: Optional[List[Leg]]
    
    def __init__(self, api_resp_dict: Dict):
        super().__init__(api_resp_dict)
        for k, v in api_resp_dict.items():
            k_upper = k.upper()
            if k_upper == 'CLASS':
                self.class_type = v
            elif k_upper == 'LEG':
                self.leg = []
                for cur_leg in v:
                    self.leg.append(Leg(cur_leg))
            else:
                setattr(self, k, v)


from dataclasses import dataclass
from typing import Dict, Optional

from fast_tradier.models.DataClassModelBase import DataClassModelBase

@dataclass
class Cash(DataClassModelBase):
    cash_available: float
    sweep: int
    unsettled_funds: float
    def __init__(self, api_resp_dict: Dict):
        super().__init__(api_resp_dict)

@dataclass
class Margin(DataClassModelBase):
    fed_call: int
    maintenance_call: int
    option_buying_power: float
    stock_buying_power: float
    stock_short_value: int
    sweep: int
    def __init__(self, api_resp_dict: Dict):
        super().__init__(api_resp_dict)

@dataclass
class Pdt(DataClassModelBase):
    fed_call: int
    maintenance_call: int
    option_buying_power: float
    stock_buying_power: float
    stock_short_value: int
    def __init__(self, api_resp_dict: Dict):
        super().__init__(api_resp_dict)

@dataclass
class AccountBalance(DataClassModelBase):
    option_short_value: int
    total_equity: float
    account_number: str
    account_type: str
    close_pl: float
    current_requirement: float
    equity: int
    long_market_value: float
    market_value: float
    open_pl: float
    option_long_value: float
    option_requirement: int
    pending_orders_count: int
    short_market_value: int
    stock_long_value: float
    total_cash: float
    uncleared_funds: int
    pending_cash: int
    margin: Optional[Margin]
    cash: Optional[Cash]
    pdt: Optional[Pdt]
    
    def __init__(self, api_resp_dict: Dict):
        self.__resp_dict = api_resp_dict
        self.margin = None
        self.cash = None
        self.pdt = None
        for k, v in api_resp_dict.items():
            key_upper = k.upper()
            if key_upper == 'MARGIN':
                self.margin = Margin(v)
            elif key_upper == 'CASH':
                self.cash = Cash(v)
            elif key_upper == 'PDT':
                self.pdt = Pdt(v)
            else:
                setattr(self, k, v)


from enum import Enum

class EquityPriceType(Enum):
    Market = "market"
    Limit = "limit"
    Stop = "stop"
    StopLimit = "stop_limit"

class OptionPriceType(Enum):
    Market = "market"
    Limit = "limit"
    Stop = "stop"
    StopLimit = "stop_limit"
    
    Debit = "debit"
    Credit = "credit"
    Even = "even"


from dataclasses import dataclass
from typing import Dict

from fast_tradier.models.DataClassModelBase import DataClassModelBase

@dataclass
class Hlocv(DataClassModelBase):
    date: str
    open: float
    high: float
    low: float
    close: float
    volume: int
    
    def __init__(self, api_resp_dict: Dict):
        super().__init__(api_resp_dict)


# flake8: noqa
from __future__ import annotations

import os
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from fast_tradier.models.DataClassModelBase import DataClassModelBase
from fast_tradier.models.trading.Duration import Duration
from fast_tradier.models.trading.OptionOrder import OptionLeg, OptionOrder
from fast_tradier.models.trading.PriceTypes import OptionPriceType
from fast_tradier.models.trading.Sides import OptionOrderSide
from fast_tradier.models.trading.TOSTradierConverter import TOSTradierConverter

short_window = int(os.environ.get("SHORT_WINDOW", "8"))
long_window = int(os.environ.get("LONG_WINDOW", "21"))


class MarketDirection(Enum):
    TrendingUp = 1
    TrendingDown = 2
    RangeBound = 3


class BuySellSignal(Enum):
    Buy = 1
    Sell = 2
    StayPut = 3


class OptionType(Enum):
    Put = 1
    Call = 2


class OptionStrategy(Enum):
    CreditCallSpread = 1
    ATMCreditCallSpread = 2

    CreditPutSpread = 3
    ATMCreditPutSpread = 4

    CreditCallSingleLeg = 5
    ATMCreditCallSingleLeg = 6

    CreditPutSingleLeg = 7
    ATMCreditPutSingleLeg = 8

    DebitCallSpread = 9
    ATMDebitCallSpread = 10

    DebitPutSpread = 11
    ATMDebitPutSpread = 12

    DebitCallSingleLeg = 13
    ATMDebitCallSingleLeg = 14

    DebitPutSingleLeg = 15
    ATMDebitPutSingleLeg = 16

    ButterflyCall = 17
    ButterflyPut = 18

    CreditBWButterflyCall = 19
    CreditBWButterflyPut = 20
    DebitBWButterflyCall = 21
    DebitBWButterflyPut = 22
    IronButterfly = 23

    """return a map of matched option strategies to that we don't duplicate the buying power used"""
    # Dict[str, Union[str, List[str]]]
    @classmethod
    def matched_option_strategies_map(
        cls,
    ) -> Dict[OptionStrategy, List[OptionStrategy]]:
        return {
            OptionStrategy.CreditCallSpread: [
                OptionStrategy.CreditPutSpread,
                OptionStrategy.ATMCreditPutSpread,
            ],
            OptionStrategy.ATMCreditCallSpread: [
                OptionStrategy.ATMCreditPutSpread,
                OptionStrategy.CreditPutSpread,
            ],
            OptionStrategy.CreditPutSpread: [
                OptionStrategy.CreditCallSpread,
                OptionStrategy.ATMCreditCallSpread,
            ],
            OptionStrategy.ATMCreditPutSpread: [
                OptionStrategy.ATMCreditCallSpread,
                OptionStrategy.ATMCreditCallSpread,
            ],
        }

    @classmethod
    def credit_spread_strategies(cls) -> List[OptionStrategy]:
        return [
            OptionStrategy.CreditCallSpread,
            OptionStrategy.CreditPutSpread,
            OptionStrategy.ATMCreditCallSpread,
            OptionStrategy.ATMCreditPutSpread,
        ]

    @classmethod
    def debit_spread_strategies(cls) -> List[OptionStrategy]:
        return [
            OptionStrategy.DebitCallSpread,
            OptionStrategy.DebitPutSpread,
            OptionStrategy.ATMDebitCallSpread,
            OptionStrategy.ATMDebitPutSpread,
        ]

    @classmethod
    def debit_butterfly_strategies(cls) -> List[OptionStrategy]:
        return [
            OptionStrategy.ButterflyPut,
            OptionStrategy.ButterflyCall,
            OptionStrategy.DebitBWButterflyCall,
            OptionStrategy.DebitBWButterflyPut,
        ]

    @classmethod
    def credit_butterfly_strategies(cls) -> List[OptionStrategy]:
        # not including IronButterFly, because it has even number of legs
        return [
            OptionStrategy.CreditBWButterflyCall,
            OptionStrategy.CreditBWButterflyPut,
        ]

    @classmethod
    def credit_single_leg_strategies(cls) -> List[OptionStrategy]:
        return [
            OptionStrategy.ATMCreditCallSingleLeg,
            OptionStrategy.ATMCreditPutSingleLeg,
            OptionStrategy.CreditCallSingleLeg,
            OptionStrategy.CreditPutSingleLeg,
        ]

    @classmethod
    def debit_single_leg_strategies(cls) -> List[OptionStrategy]:
        return [
            OptionStrategy.DebitPutSingleLeg,
            OptionStrategy.DebitCallSingleLeg,
            OptionStrategy.ATMDebitCallSingleLeg,
            OptionStrategy.ATMDebitPutSingleLeg,
        ]


class OptionStrikeInfo(object):
    def __init__(
        self,
        underlying_symbol: str,
        option_symbol: str,
        strike: float,
        bid: float,
        ask: float,
        last_price: float,
        option_type: OptionType,
        convert_tos_symbol: bool = True,
    ) -> None:
        self.__underlying_symbol = underlying_symbol
        self.__option_symbol = (
            TOSTradierConverter.tos_to_tradier(option_symbol)
            if convert_tos_symbol
            else option_symbol
        )
        self.__tos_option_symbol = (
            TOSTradierConverter.tradier_to_tos(option_symbol)
            if convert_tos_symbol
            else option_symbol
        )
        self.__strike = strike
        self.__bid = bid
        self.__ask = ask
        self.__last_price = last_price
        self.__option_type = option_type

    @property
    def option_symbol(self) -> str:
        return self.__option_symbol

    @property
    def tos_option_symbol(self) -> str:
        return self.__tos_option_symbol

    @property
    def strike(self) -> float:
        return self.__strike

    @property
    def bid(self) -> float:
        return self.__bid

    @property
    def ask(self) -> float:
        return self.__ask

    @property
    def last_price(self) -> float:
        return self.__last_price

    @property
    def option_type(self) -> OptionType:
        return self.__option_type


@dataclass
class TradierOptionOrderMetadata(DataClassModelBase):
    lower_bound: Optional[float]
    upper_bound: Optional[float]
    predicted_close: Optional[float]
    db_primary_id: Optional[int]

    def __init__(self, json_obj: Dict) -> None:  # type: ignore
        super().__init__(json_obj)


class SimpleOptionChainPair(object):
    def __init__(
        self,
        underlying_symbol: str,
        call_df: pd.DataFrame,
        put_df: pd.DataFrame,
    ) -> None:
        self.__underlying_symbol = underlying_symbol
        self.__call_df = call_df
        self.__put_df = put_df

    @property
    def underlying_symbol(self) -> str:
        return self.__underlying_symbol

    @property
    def call_df(self) -> pd.DataFrame:
        return self.__call_df

    @property
    def put_df(self) -> pd.DataFrame:
        return self.__put_df

    """
        find the latest price from option chain for the option_symbol
        Note: option_symbol could be either TOS or Tradier
        returns bid, ask, lastPrice
    """

    def find_price(self, option_symbol: str) -> Tuple:  # type: ignore
        try:
            option_symbol = option_symbol.upper()
            iloc_idx_call = self.call_df.loc[self.call_df["symbol"] == option_symbol]
            iloc_idx_put = self.put_df.loc[self.put_df["symbol"] == option_symbol]
            target_idx = -1
            target_df = None
            result = (-1, -1, -1)

            if len(iloc_idx_call.index) > 0:
                target_idx = self.call_df.index.get_loc(iloc_idx_call.index[0])
                target_df = self.call_df
            elif len(iloc_idx_put.index) > 0:
                target_idx = self.put_df.index.get_loc(iloc_idx_put.index[0])
                target_df = self.put_df
            else:
                raise Exception("option symbol not found")
        except Exception as ex:
            print(f"exception getting price for option symbol {option_symbol}: {ex} ")

        if target_df is not None:
            bid = target_df.iloc[target_idx]["bid"]
            ask = target_df.iloc[target_idx]["ask"]
            lastPrice = target_df.iloc[target_idx]["lastPrice"]
            if lastPrice == 0:
                # when lastPrice is 0, try to average bid and ask
                lastPrice = np.round(np.mean(bid, ask), 2)
            result = (bid, ask, lastPrice)
        return result


class TradierOptionOrder(OptionOrder):
    def __init__(
        self,
        ticker: str,
        price: float,
        price_type: OptionPriceType,
        duration: Duration,
        option_legs: List[OptionLeg],
        timestamp: Optional[int] = None,
        stop_loss_factor: float = 0.5,
        stop_win_factor: float = 0.8,
        option_strategy: OptionStrategy = OptionStrategy.CreditCallSpread,
        init_stop_loss_hits: int = 0,
        close_status: Optional[str] = None,
        close_order_id: Optional[int] = None,
        meta_data: Optional[TradierOptionOrderMetadata] = None,
    ):
        super().__init__(ticker, price, price_type, duration, option_legs)
        # self.__buying_power_used = price # TODO: needs to be updated by OrderManager, #FIXME: use the actual amount, not price
        self.__timestamp = (
            int(datetime.now().timestamp()) if timestamp is None else timestamp
        )
        self.__stop_loss_factor = stop_loss_factor
        self.__stop_win_factor = stop_win_factor
        self.__option_strategy = option_strategy
        self.__stop_loss_hit_count = init_stop_loss_hits
        self.__close_status = close_status
        self.__close_order_id = close_order_id
        self.__meta_data = meta_data

    @property
    def close_status(self) -> Optional[str]:
        return self.__close_status

    @close_status.setter
    def close_status(self, new_val: str) -> None:
        self.__close_status = new_val

    @property
    def close_order_id(self) -> Optional[int]:
        return self.__close_order_id

    @close_order_id.setter
    def close_order_id(self, new_val: Optional[int]) -> None:
        self.__close_order_id = new_val

    @property
    def metadata(self) -> Optional[TradierOptionOrderMetadata]:
        return self.__meta_data

    @metadata.setter
    def metadata(self, new_value: Dict) -> None:  # type: ignore
        if new_value is not None:
            self.__meta_data = TradierOptionOrderMetadata(new_value)

    # TODO: rename to buying_power_required ?
    @property
    def buying_power_used(self) -> float:
        return self.__buying_power_used

    @buying_power_used.setter
    def buying_power_used(self, new_value: float) -> None:
        self.__buying_power_used = new_value

    @property
    def is_buying_power_released(self) -> bool:
        return (
            self.status
            not in ["open", "pending", "closing", "filled", "partially_filled"]
            or self.close_status == "filled"
        )

    @property
    def waiting_for_fill(self) -> bool:
        return self.status in ["open", "partially_filled", "pending"]

    @property
    def is_filled(self) -> bool:
        return self.status in ["filled", "partially_filled"]

    @property
    def is_closed(self) -> bool:
        return self.close_status == "filled" and self.close_order_id is not None

    @property
    def close_status_needs_confirm(self) -> bool:
        return (
            self.status == "closing"
            and self.close_status != "filled"
            and self.close_order_id is not None
        )

    @property
    def is_open(self) -> bool:
        return self.status.upper() == "OPEN"

    @property
    def is_pending(self) -> bool:
        return self.status.upper() == "PENDING"

    @property
    def is_closing(self) -> bool:
        return self.status.upper() == "CLOSING"

    @property
    def timestamp(self) -> int:
        return self.__timestamp

    @property
    def stop_loss_factor(self) -> float:
        return self.__stop_loss_factor

    @property
    def stop_win_factor(self) -> float:
        return self.__stop_win_factor

    @property
    def option_strategy(self) -> OptionStrategy:
        return self.__option_strategy

    @property
    def stop_loss_hit_count(self) -> int:
        return self.__stop_loss_hit_count

    def increment_stop_loss_hit(self) -> None:
        self.__stop_loss_hit_count += 1

    def tradier_option_symbols(self) -> List[str]:
        return [opt_leg.option_symbol for opt_leg in self.option_legs]

    def flip_price_type(self) -> str:
        """reverse the price_type from credit to debit or vice versa"""
        if self.price_type.lower() == "credit":
            return "debit"
        elif self.price_type.lower() == "debit":
            return "credit"

        return self.price_type

    """
    given the latest call and put option chain. calculates
    the profit amount and its percentage in buying power used
    tuple 0: profit amount, 1: profit vs buying power used
    """

    def profit_info(self, call_df: pd.DataFrame, put_df: pd.DataFrame) -> Tuple:  # type: ignore
        # FIXME: what if the number of legs are not the same for each option symbol. e.g. Butterfly ?
        contract_count = (
            1
            if self.option_legs is None or len(self.option_legs) == 0
            else self.option_legs[0].quantity
        )
        cur_val = 0
        profit_bp_percentage = 0

        # find the option curr price in the dataframe, and calculate the profit vs the price.
        opt_chain_pair = SimpleOptionChainPair(
            underlying_symbol=self.ticker,
            call_df=call_df,
            put_df=put_df,
        )
        min_leg_count = 1
        for opt_leg in self.option_legs:
            min_leg_count = min(min_leg_count, opt_leg.quantity)
            bid, ask, lastPrice = opt_chain_pair.find_price(
                opt_leg.tos_option_symbol,
            )  # TODO: try tos option symbol first, because it queries TOS option chain # FIXME with regex
            if lastPrice == -1:
                # Try tradier option as a backup
                bid, ask, lastPrice = opt_chain_pair.find_price(
                    opt_leg.option_symbol,
                )  # TODO: try both TOS and Tradier option symbols?

            if lastPrice > -1:
                if opt_leg.side == "sell_to_open":
                    cur_val += (
                        lastPrice * opt_leg.quantity
                    )  ### * 100 ??? # FIXME: what if the number of option legs are different int he order ???
                elif opt_leg.side == "buy_to_open":
                    cur_val -= lastPrice * opt_leg.quantity

        profit = 0.0
        if (
            self.option_strategy in OptionStrategy.debit_spread_strategies()
            or self.option_strategy in OptionStrategy.debit_single_leg_strategies()
        ):
            contract_count = self.option_legs[0].quantity
            profit = cur_val - (self.price * contract_count)
        elif (
            self.option_strategy in OptionStrategy.credit_spread_strategies()
            or self.option_strategy
            == self.option_strategy
            == OptionStrategy.IronButterfly
        ):
            contract_count = self.option_legs[0].quantity
            profit = (self.price * contract_count) - cur_val
        elif self.option_strategy in OptionStrategy.debit_butterfly_strategies():
            contract_count = min_leg_count
            profit = cur_val - (self.price * contract_count)
        elif self.option_strategy in OptionStrategy.credit_butterfly_strategies():
            contract_count = min_leg_count
            profit = (self.price * contract_count) - cur_val

        # TODO: make 100.0 a constant
        profit_bp_percentage = profit * 100.0 / self.buying_power_used  # type: ignore
        # profit per contract, needs to be divided by contract_count.
        # FIXME: but what happens to butterfly strategies?
        return np.round(profit / (1.0 * contract_count), 3), np.round(
            profit_bp_percentage,
            3,
        )

    def to_json(self) -> Dict:  # type: ignore
        result = super().to_json()
        # result = {
        #     "id": self.id,
        #     "status": self.status,
        #     "symbol": self.ticker,
        #     "duration": self.duration,
        #     "price": self.price,
        #     "type": self.price_type,
        #     "class": self.order_class,
        # }
        if self.metadata is not None:
            result["metadata"] = self.metadata.to_json()
        return result

    # def __str__(self) -> str:
    #     legs = ""
    #     for i in range(len(self.option_legs)):
    #         opt_leg = self.option_legs[i]
    #         legs += f"leg[{i}]:{opt_leg.option_symbol}({opt_leg.side});"
    #     return f"id:{self.id},price:{self.price}|strategy:{self.option_strategy.name}|legs:[{legs}]"

if __name__ == '__main__':
    ticker = "spx"
    leg_quantity = 1
    price = 1.20
    option_leg1 = OptionLeg(
        underlying_symbol=ticker,
        option_symbol="SPXW_052223C4225",
        side=OptionOrderSide.SellToOpen,
        quantity=leg_quantity,
    )
    option_leg2 = OptionLeg(
        underlying_symbol=ticker,
        option_symbol="SPXW_052223C4235",
        side=OptionOrderSide.BuyToOpen,
        quantity=leg_quantity,
    )
    option_legs = [
        option_leg1,
        option_leg2,
    ]
    stop_loss_factor = 3.0
    stop_win_facotr = 0.5
    meta_data = TradierOptionOrderMetadata(
        {
            "lower_bound": 4100,
            "upper_bound": 4200,
            "predicted_close": 4150,
        },
    )

    order = TradierOptionOrder(
        ticker=ticker,
        price=price,
        price_type=OptionPriceType.Market,
        duration=Duration.Day,
        option_legs=option_legs,
        timestamp=int(datetime.now().timestamp()),
        stop_loss_factor=stop_loss_factor,
        stop_win_factor=stop_win_facotr,
        option_strategy=OptionStrategy.CreditCallSpread,
        close_status='pending',
        close_order_id=100,
        meta_data=meta_data,
    )
    
    print('order.to_json: ', order.to_json())


from dataclasses import dataclass
from typing import Dict

from fast_tradier.models.DataClassModelBase import DataClassModelBase

@dataclass
class Position(DataClassModelBase):
    cost_basis: float
    date_acquired: str
    id: int
    quantity: float
    symbol: str
    
    def __init__(self, api_resp_dict: Dict):
        super().__init__(api_resp_dict)